# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"totensor": "000_utils.ipynb",
         "toarray": "000_utils.ipynb",
         "to3dtensor": "000_utils.ipynb",
         "to2dtensor": "000_utils.ipynb",
         "to1dtensor": "000_utils.ipynb",
         "to3darray": "000_utils.ipynb",
         "to2darray": "000_utils.ipynb",
         "to1darray": "000_utils.ipynb",
         "to3d": "000_utils.ipynb",
         "to2d": "000_utils.ipynb",
         "to1d": "000_utils.ipynb",
         "to2dPlus": "000_utils.ipynb",
         "to3dPlus": "000_utils.ipynb",
         "to2dPlusTensor": "000_utils.ipynb",
         "to2dPlusArray": "000_utils.ipynb",
         "to3dPlusTensor": "000_utils.ipynb",
         "to3dPlusArray": "000_utils.ipynb",
         "Todtype": "000_utils.ipynb",
         "bytes2size": "000_utils.ipynb",
         "bytes2GB": "000_utils.ipynb",
         "delete_all_in_dir": "000_utils.ipynb",
         "reverse_dict": "000_utils.ipynb",
         "is_tuple": "000_utils.ipynb",
         "itemify": "000_utils.ipynb",
         "is_none": "000_utils.ipynb",
         "ifisnone": "000_utils.ipynb",
         "ifnoneelse": "000_utils.ipynb",
         "ifisnoneelse": "000_utils.ipynb",
         "ifelse": "000_utils.ipynb",
         "is_not_close": "000_utils.ipynb",
         "test_not_close": "000_utils.ipynb",
         "test_type": "000_utils.ipynb",
         "stack": "000_utils.ipynb",
         "cat2int": "000_utils.ipynb",
         "cycle_dl": "000_utils.ipynb",
         "memmap2cache": "000_utils.ipynb",
         "TrainValTestSplitter": "000b_data.validation.ipynb",
         "get_splits": "000b_data.validation.ipynb",
         "check_overlap": "000b_data.validation.ipynb",
         "leakage_finder": "000b_data.validation.ipynb",
         "get_predefined_splits": "000b_data.validation.ipynb",
         "combine_split_data": "000b_data.validation.ipynb",
         "decompress_from_url": "001_data.external.ipynb",
         "get_UCR_univariate_list": "001_data.external.ipynb",
         "get_UCR_multivariate_list": "001_data.external.ipynb",
         "stack_padding": "001_data.external.ipynb",
         "get_UCR_data": "001_data.external.ipynb",
         "NumpyTensor": "002_data.core.ipynb",
         "ToNumpyTensor": "002_data.core.ipynb",
         "TSTensor": "002_data.core.ipynb",
         "ToTSTensor": "002_data.core.ipynb",
         "NumpyTensorBlock": "002_data.core.ipynb",
         "TSTensorBlock": "002_data.core.ipynb",
         "TorchDataset": "002_data.core.ipynb",
         "NumpyDataset": "002_data.core.ipynb",
         "TSDataset": "002_data.core.ipynb",
         "NumpyDatasets": "002_data.core.ipynb",
         "TSDatasets": "002_data.core.ipynb",
         "add_ds": "002_data.core.ipynb",
         "NumpyDatasets.add_test": "002_data.core.ipynb",
         "NumpyDatasets.add_unlabeled": "002_data.core.ipynb",
         "NumpyDataLoader": "002_data.core.ipynb",
         "show_tuple": "002_data.core.ipynb",
         "TSDataLoader": "002_data.core.ipynb",
         "NumpyDataLoaders": "002_data.core.ipynb",
         "TSDataLoaders": "002_data.core.ipynb",
         "DataLoader.cws": "002_data.core.ipynb",
         "TSStandardize": "003_data.transforms.ipynb",
         "torch.Tensor.mul_min": "003_data.transforms.ipynb",
         "TSTensor.mul_min": "003_data.transforms.ipynb",
         "NumpyTensor.mul_min": "003_data.transforms.ipynb",
         "torch.Tensor.mul_max": "003_data.transforms.ipynb",
         "TSTensor.mul_max": "003_data.transforms.ipynb",
         "NumpyTensor.mul_max": "003_data.transforms.ipynb",
         "TSNormalize": "003_data.transforms.ipynb",
         "TSIdentity": "003_data.transforms.ipynb",
         "TSShuffle_HLs": "003_data.transforms.ipynb",
         "TSMagNoise": "003_data.transforms.ipynb",
         "TSMagMulNoise": "003_data.transforms.ipynb",
         "random_curve_generator": "003_data.transforms.ipynb",
         "random_cum_curve_generator": "003_data.transforms.ipynb",
         "random_cum_noise_generator": "003_data.transforms.ipynb",
         "TSTimeNoise": "003_data.transforms.ipynb",
         "TSMagWarp": "003_data.transforms.ipynb",
         "TSTimeWarp": "003_data.transforms.ipynb",
         "TSMagScale": "003_data.transforms.ipynb",
         "TSMagScaleVar": "003_data.transforms.ipynb",
         "TSZoomIn": "003_data.transforms.ipynb",
         "TSZoomOut": "003_data.transforms.ipynb",
         "TSScale": "003_data.transforms.ipynb",
         "TSRandomTimeStep": "003_data.transforms.ipynb",
         "TSBlur": "003_data.transforms.ipynb",
         "TSSmooth": "003_data.transforms.ipynb",
         "maddest": "003_data.transforms.ipynb",
         "TSDenoise": "003_data.transforms.ipynb",
         "TSRandomNoise": "003_data.transforms.ipynb",
         "TSLookBack": "003_data.transforms.ipynb",
         "TSVarOut": "003_data.transforms.ipynb",
         "TSCutOut": "003_data.transforms.ipynb",
         "TSTimeStepOut": "003_data.transforms.ipynb",
         "TSCrop": "003_data.transforms.ipynb",
         "TSRandomCrop": "003_data.transforms.ipynb",
         "TSRandomResizedCrop": "003_data.transforms.ipynb",
         "TSCenterCrop": "003_data.transforms.ipynb",
         "TSMaskOut": "003_data.transforms.ipynb",
         "TSTranslateX": "003_data.transforms.ipynb",
         "TSFlip": "003_data.transforms.ipynb",
         "TSRandomFlip": "003_data.transforms.ipynb",
         "TSShift": "003_data.transforms.ipynb",
         "TSRandomRotate": "003_data.transforms.ipynb",
         "TSNeg": "003_data.transforms.ipynb",
         "TSRandomNeg": "003_data.transforms.ipynb",
         "TSFreqNoise": "003_data.transforms.ipynb",
         "TSFreqWarp": "003_data.transforms.ipynb",
         "TSFreqScale": "003_data.transforms.ipynb",
         "TabularDataset": "005_data.tabular.ipynb",
         "TabularDataLoader": "005_data.tabular.ipynb",
         "MatthewsCorrCoef": "007_metrics.ipynb",
         "Learner.save_all": "008_learner.ipynb",
         "load_learner_all": "008_learner.ipynb",
         "Recorder.plot_metrics": "008_learner.ipynb",
         "LookAhead": "009_optimizer.ipynb",
         "Ralamb": "009_optimizer.ipynb",
         "ralamb": "009_optimizer.ipynb",
         "RangerLars": "009_optimizer.ipynb",
         "rangerlars": "009_optimizer.ipynb",
         "Ranger": "009_optimizer.ipynb",
         "rangergc": "009_optimizer.ipynb",
         "generate_kernels": "010_rocket_functions.ipynb",
         "apply_kernel": "010_rocket_functions.ipynb",
         "apply_kernels": "010_rocket_functions.ipynb",
         "ROCKET": "010_rocket_functions.ipynb",
         "noop": "100_layers.ipynb",
         "mish": "100_layers.ipynb",
         "Mish": "100_layers.ipynb",
         "get_act_layer": "100_layers.ipynb",
         "same_padding1d": "100_layers.ipynb",
         "Pad1d": "100_layers.ipynb",
         "Conv1dSame": "100_layers.ipynb",
         "Chomp1d": "100_layers.ipynb",
         "Conv1dCausal": "100_layers.ipynb",
         "Conv1d": "100_layers.ipynb",
         "CoordConv1D": "100_layers.ipynb",
         "LambdaPlus": "100_layers.ipynb",
         "Flatten": "100_layers.ipynb",
         "Squeeze": "100_layers.ipynb",
         "Unsqueeze": "100_layers.ipynb",
         "YRange": "100_layers.ipynb",
         "Temp": "100_layers.ipynb",
         "get_layers": "100b_models_utils.ipynb",
         "count_params": "100b_models_utils.ipynb",
         "ResBlock": "101_ResNet.ipynb",
         "ResNet": "101_ResNet.ipynb",
         "shortcut": "102_InceptionTime.ipynb",
         "Inception": "102_InceptionTime.ipynb",
         "InceptionBlock": "102_InceptionTime.ipynb",
         "InceptionTime": "102_InceptionTime.ipynb",
         "FCN": "103_FCN.ipynb",
         "Block": "104_ResCNN.ipynb",
         "ResCNN": "104_ResCNN.ipynb"}

modules = ["utils.py",
           "data/validation.py",
           "data/external.py",
           "data/core.py",
           "data/transforms.py",
           "data/tabular.py",
           "metrics.py",
           "learner.py",
           "optimizer.py",
           "rocket_functions.py",
           "models/layers.py",
           "models/utils.py",
           "models/ResNet.py",
           "models/InceptionTime.py",
           "models/FCN.py",
           "models/ResCNN.py"]

doc_url = "https://timeseriesai.github.io/timeseriesAI//"

git_url = "https://github.com/timeseriesAI/timeseriesAI/tree/master/"

def custom_doc_links(name): return None
